//04_09_19 Joerg Homberger

/* 
 * Implementierung eines Genetischen Algorithmus zur Lsung des Standortproblems
 * Die zu lsende Probleminstanz wird zu Beginn des Algorithmus als Beispiel zufaellig erzeugt
 * Alternativ knnen auch vorgegebenen Instanzen aus der Literatur berechnet werden
 * Hierzu ist die zu lsende Instanz anstelle der Erzeugung einer Instanz einzulesen
 * Beispiele fr Instanzen finden sich z.B. unter dem Begriff "Uncapacitated facility location problem": http://resources.mpi-inf.mpg.de/departments/d1/projects/benchmarks/UflLib/ (22.09.19)
 * Selektion: Wettkampf
 * Ersetzung: Generational Replacement (nicht-elitr)
 * Mutation:  Flip-Mutation
 * Crossover: Ein-Punkt-Crossover
 * Viel Spass beim Testen und Weiterentwickeln ;-) euer Jrg Homberber: joerg.homberger@hft-stuttgart.de
 * 
 * 
 */

public class Ablauf {
	// Parameter des Genetischen Algorithmus
	static int anz   = 20;  				//Anzahl Individuen
	static int maxit = 1000;				//Maximale Anzahl an Itertionen
	static boolean minimierung = true;		//Minimierungsproblem
	
	// Datenstrukturen fr GA
	static int[][] POP = new int[anz][];	  //Population aus anz Individuen
	static int[]   FIT = new int[anz];		  //Fitness fr anz Individuen
	static int[][] POPCHILD = new int[anz][]; //Kinder aus anz Individuen
	static int[]   FITCHILD = new int[anz];	  //Fitness fr anz Kinder
	static int[]   BEST;	                  //Beste lsung bzw. Ergebnis
	
	
	
	// Variablen fr GA
	static int iteration;					//Iterationszaehler
	static int bestFit;						//Beste Fitness 
	
	public static void main(String[] args) {
		if(minimierung)	bestFit = Integer.MAX_VALUE; 
		else 			bestFit = Integer.MIN_VALUE; 

		//Problem.einlesenInstanz("123UnifS.txt");
		Problem.erzeugenInstanz(100, 100, 100);		
		Problem.initialisierung();
		bewertung(POP,FIT);
		
		for (iteration = 1; iteration < maxit; iteration++) {
			fortpflanzung();
			bewertung(POPCHILD,FITCHILD);
			ersetzung();
		}
		Problem.loesungAusgeben(BEST);

		System.out.println(Problem.fitness(BEST));
	}

	///////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////
	public static void fortpflanzung() {
		int child1 = 0;
		int child2 = 1;
		int anzahlCrossover = anz/2;
		
		for (int i = 0; i < anzahlCrossover; i++) {
			//Selektion der Eltern
			//Zweigeschlechtliche Fortpflanzung
			int eltern1 = selektionEltern();
			int eltern2 = selektionEltern();
			while(eltern1 == eltern2) eltern2 = selektionEltern();

			Problem.crossover(child1, child2, eltern1, eltern2);
			Problem.flipMutation(child1);
			Problem.flipMutation(child2);
			
			child1 += 2;
			child2 += 2;
		}
		if(anz%2 != 0){
			//fr den Fall, dass Populationsgre eine ungerade Zahl ist
			//das Child POPCHILD.length-1 wurde noch nicht erzeugt
			//"Eingeschlechtliche" Fortpflanzung
			int eltern = selektionEltern();
			int child  = POPCHILD.length-1;
			Problem.crossover(child, child, eltern, eltern);
			Problem.flipMutation(child);
		}
	}

	public static int selektionEltern(){
		//Wettkampf-Selektion
		int eltern;
		int kandidat1 = (int) (POP.length * Math.random());
		int kandidat2 = (int) (POP.length * Math.random());
		if(minimierung){
			if(FIT[kandidat1] < FIT[kandidat2])eltern = kandidat1;
			else                               eltern = kandidat2;
		}
		else{
			if(FIT[kandidat1] > FIT[kandidat2])eltern = kandidat1;
			else                               eltern = kandidat2;
		}
		return eltern;
	}
	
	///////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////
	public static void bewertung(int[][] ind, int[] fit) {
		int aktFit;
		if(minimierung)aktFit = Integer.MAX_VALUE;
		else           aktFit = Integer.MIN_VALUE;
		
		for (int i = 0; i < fit.length; i++) {
			fit[i] = Problem.fitness(ind[i]);
			
			//System.out.println(i + " Fit " + fit[i]);
			
			if(minimierung){
				if (fit[i] < bestFit) {
					bestFit = fit[i];
					for(int j=0;j<POP[i].length;j++){
						BEST[j] = POP[i][j];
					}
				}
				if (fit[i] < aktFit) {
					aktFit = fit[i];
				}
			}
			else{
				if (fit[i] > bestFit) {
					bestFit = fit[i];
					for(int j=0;j<POP[i].length;j++){
						BEST[j] = POP[i][j];
					}
				}
				if (fit[i] > aktFit) {
					aktFit = fit[i];
				}
			}
		}
		System.out.println("Iteration: " + iteration + " Aktuelle Fitnes: " + aktFit + " Beste Fitnes: " + bestFit);

	}

	public static void ausgabePop() {
		System.out.println("Eltern:");
		for (int indi = 0; indi < POP.length; indi++) {
			System.out.print(indi + " ");
			for (int st = 0; st < POP[indi].length; st++) {
				System.out.print(POP[indi][st]);
			}
			System.out.print(" " + FIT[indi]);
			System.out.println();
		}
	}
	public static void ausgabeChild() {

		System.out.println("Childs:");
		for (int indi = 0; indi < POPCHILD.length; indi++) {
			System.out.print(indi + " ");
			for (int st = 0; st < POPCHILD[indi].length; st++) {
				System.out.print(POPCHILD[indi][st]);
			}
			System.out.print(" " + FITCHILD[indi]);
			System.out.println();
		}
	}
	

	
	
	
	///////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////

	public static void ersetzung() {
		//GenerationalReplacement
		int bestPopFit   = Integer.MAX_VALUE;
		int bestChildFit = Integer.MAX_VALUE;
		int bestPop   = 0;
		int bestChild = 0;
		boolean verschlechterung = false;
		
		for(int i=0;i<POP.length;i++){
			if(FIT[i]<bestPopFit){
				bestPopFit = FIT[i];
				bestPop    = i;
			}
		}
		for(int i=0;i<POPCHILD.length;i++){
			if(FITCHILD[i]<bestChildFit){
				bestChildFit = FITCHILD[i];
				bestChild    = i;
			}
		}
		if(minimierung){
			if(bestChildFit > bestPopFit)verschlechterung = true;	
		}
		else{
			if(bestChildFit < bestPopFit)verschlechterung = true;
		}
		//System.out.println(bestPopFit + " " + bestChildFit + " " + verschlechterung);
		
		////////////////////////////////////////////////////////
		
		for(int i=0;i<POP.length;i++) {
			for(int s=0;s<POP[i].length;s++) {
					POP[i][s] = POPCHILD[i][s];
			}
			FIT[i] = FITCHILD[i];
		}		
	}
	
	
	


}
